/*10:*/
#line 518 "dictionary.web"

#include <fork.h>
#include <io.h>
#include "pad.h"
#include "prefix.h"
#include "merge.h"

/*11:*/
#line 569 "dictionary.web"

#define MAX_CHILDREN 3

typedef struct dict_node{
int key;
int no_of_children;
int no_of_leaves;
int child_no;
int utility;
struct dict_node*children[MAX_CHILDREN];
struct dict_node*father;
}DICT_NODE;

/*:11*//*15:*/
#line 634 "dictionary.web"

#define dict_size(dictionary) (dictionary)->no_of_leaves
#define dict_empty(dictionary) ((dictionary)->no_of_leaves==0)

/*:15*/
#line 525 "dictionary.web"

/*7:*/
#line 450 "dictionary.web"

extern async DICT_NODE*dictitem_new(pr int key,pr void*content);
extern async int dictitem_key(pr DICT_NODE*item);
extern async void*dictitem_newcontent(pr DICT_NODE*item,pr void*content);


extern sync DICT_NODE*dict_build(sh DICT_NODE*item[],sh int n);
extern sync int dict_flatten(sh DICT_NODE*dictionary,sh DICT_NODE*item[]);
extern async DICT_NODE*dict_closest(pr DICT_NODE*dictionary,pr int key);
extern async DICT_NODE*dict_search(pr DICT_NODE*dictionary,pr int key);
extern async DICT_NODE*dict_retrieve(pr DICT_NODE*dictionary,pr int k);
extern async DICT_NODE*dict_find(pr DICT_NODE*dictionary,
pr DICT_NODE*searchitem,
async DICT_NODE*(*select)(pr DICT_NODE*,
pr DICT_NODE*));
extern async int dict_sum(pr DICT_NODE*item,
async int(*treesum)(pr DICT_NODE*));
extern sync DICT_NODE*dict_insert(sh DICT_NODE*dictionary,
sh DICT_NODE*item[],sh int n);
extern sync DICT_NODE*dict_delete(sh DICT_NODE*dictionary,
sh DICT_NODE*oldnodes[],sh int n);

/*:7*//*14:*/
#line 623 "dictionary.web"

#define dictitem_key(item) (item)->key
#define dictitem_content(item) (item)->children[0]
#define dictitem_free(item) free(item)

/*:14*/
#line 526 "dictionary.web"

/*12:*/
#line 585 "dictionary.web"

#define LEFT children[0]
#define MIDDLE children[1]
#define RIGHT children[2]
#define CONTENT children[0] 


#define GROUP_NO utility
#define PENDING utility


/*:12*/
#line 527 "dictionary.web"


/*:10*//*13:*/
#line 599 "dictionary.web"

async DICT_NODE*dictitem_new(pr int key,pr void*content)
{
pr DICT_NODE*item;

item= (DICT_NODE*)malloc(sizeof(DICT_NODE));
item->key= key;
item->CONTENT= content;

return item;
}

#define dictitem_free(item) free(item)

async void*dictitem_newcontent(pr DICT_NODE*item,pr void*content)
{
pr void*oldcontent;

oldcontent= item->CONTENT;
item->CONTENT= content;
return oldcontent;
}

/*:13*//*16:*/
#line 663 "dictionary.web"

sync DICT_NODE*dict_build(sh DICT_NODE*item[],sh int n)
{
sh int p= groupsize();

sh int nn;
pr int i;

sh DICT_NODE**subtree;
pr DICT_NODE*node;


subtree= (DICT_NODE**)shalloc(n*sizeof(DICT_NODE*));


for(i= ME;i<n;i+= p){
item[i]->no_of_children= 0;
item[i]->no_of_leaves= 1;

subtree[i]= item[i];
}


while(n>1){
nn= n>>1;
for(i= ME;i<nn;i+= p){


node= (DICT_NODE*)malloc(sizeof(DICT_NODE));

node->LEFT= subtree[i<<1];
node->MIDDLE= subtree[(i<<1)+1];
node->key= node->MIDDLE->key;
node->no_of_children= 2;
node->no_of_leaves= 
node->LEFT->no_of_leaves+node->MIDDLE->no_of_leaves;

subtree[i]= node;

node->LEFT->child_no= 0;
node->LEFT->father= node;
node->MIDDLE->child_no= 1;
node->MIDDLE->father= node;
}
i= i-p;


if((odd(n))&&(i==nn-1)){

node->RIGHT= subtree[n-1];
node->key= subtree[n-1]->key;
node->no_of_children= 3;
node->no_of_leaves+= node->RIGHT->no_of_leaves;

node->RIGHT->child_no= 2;
node->RIGHT->father= node;
}
n= nn;
}
if(ME==0){
subtree[0]->child_no= 0;
subtree[0]->father= NULL;
}
shallfree();
return subtree[0];
}

/*:16*//*17:*/
#line 765 "dictionary.web"

sync int dict_flatten(sh DICT_NODE*dictionary,sh DICT_NODE*item[])
{
sh int p= groupsize();

pr DICT_NODE*node;
sh int*index1,*index2,*index;
sh int n,nn,nnn;
sh bool active;
pr int i,ii,nc;

n= dict_size(dictionary);
index1= (int*)shalloc(2*n*sizeof(int));
index2= index1+n;

for(i= ME;i<n;i+= p){
item[i]= NULL;
index1[i]= -1;index2[i]= -1;
}

if(ME==0){
nn= 1;
index1[0]= 0;
item[0]= dictionary;
active= item[0]->no_of_children!=0;
}
while(active){
if(ME==0)nnn= MAX_CHILDREN*nn;
for(i= ME;i<nn;i+= p){
ii= index1[i];
if(ii>=0){
node= item[ii];
index1[i]= -1;
for(nc= 0;nc<node->no_of_children;nc++){

if(nnn<n)index2[MAX_CHILDREN*i+nc]= ii;
else index2[ii]= ii;
item[ii]= node->children[nc];
ii+= node->children[nc]->no_of_leaves;
}

}
}
if(ME==0){

active= item[0]->no_of_children!=0;
if(nnn<n)nn= nnn;else nn= n;

index= index1;index1= index2;index2= index;
}
}

return n;
}

/*:17*//*18:*/
#line 845 "dictionary.web"

async DICT_NODE*dict_closest(pr DICT_NODE*dictionary,pr int key)
{
pr DICT_NODE*node;

node= dictionary;
while(node->no_of_children>0){
if(key<=node->LEFT->key)
node= node->LEFT;
else if((key<=node->MIDDLE->key)||(node->no_of_children==2))
node= node->MIDDLE;
else
node= node->RIGHT;
}
return node;
}

async DICT_NODE*dict_search(pr DICT_NODE*dictionary,pr int key)
{
pr DICT_NODE*leaf;

leaf= dict_closest(dictionary,key);
if(leaf->key!=key)leaf= NULL;
return leaf;
}

/*:18*//*20:*/
#line 896 "dictionary.web"

async DICT_NODE*dict_retrieve(pr DICT_NODE*dictionary,pr int k)
{
pr DICT_NODE*node;

if((k<1)||(k>dict_size(dictionary)))return NULL;

node= dictionary;
while(node->no_of_children>0){
if(k<=node->LEFT->no_of_leaves)node= node->LEFT;
else{
k= k-node->LEFT->no_of_leaves;
if((k<=node->MIDDLE->no_of_leaves)||(node->no_of_children==2))
node= node->MIDDLE;
else{
k= k-node->MIDDLE->no_of_leaves;
node= node->RIGHT;
}
}
}
return node;
}

/*:20*//*21:*/
#line 934 "dictionary.web"

async DICT_NODE*dict_find(pr DICT_NODE*dictionary,
pr DICT_NODE*searchitem,
async DICT_NODE*(*select)(pr DICT_NODE*,
pr DICT_NODE*))
{
pr DICT_NODE*node;

node= dictionary;
while(node->no_of_children>0)node= (*select)(searchitem,node);
return node;
}



async DICT_NODE*retrieve(pr DICT_NODE*searchitem,pr DICT_NODE*node)
{
if(searchitem->no_of_leaves<=node->LEFT->no_of_leaves)node= node->LEFT;
else{
searchitem->no_of_leaves= 
searchitem->no_of_leaves-node->LEFT->no_of_leaves;
if((searchitem->no_of_leaves<=node->MIDDLE->no_of_leaves)||
(node->no_of_children==2))
node= node->MIDDLE;
else{
searchitem->no_of_leaves= 
searchitem->no_of_leaves-node->MIDDLE->no_of_leaves;
node= node->RIGHT;
}
}
return node;
}



async DICT_NODE*dict_retrv(pr DICT_NODE*dictionary,pr int k)
{
pr DICT_NODE knode;

knode.no_of_leaves= k;
return dict_find(dictionary,&knode,retrieve);
}

/*:21*//*22:*/
#line 985 "dictionary.web"

async int dict_sum(pr DICT_NODE*item,
async int(*treesum)(pr DICT_NODE*))
{
pr int sum;

sum= 0;

while(item!=NULL){
sum+= (*treesum)(item);
item= item->father;
}
return sum;
}

sync int leaves_before(pr DICT_NODE*node)
{
pr int nc,no_of_leaves;

no_of_leaves= 0;
if(node->father!=NULL)
for(nc= 0;nc<node->child_no;nc++)
no_of_leaves+= node->father->children[nc]->no_of_leaves;
return no_of_leaves;
}

#define dict_position(item) dict_sum(item,leaves_before)

/*:22*//*23:*/
#line 1134 "dictionary.web"

/*34:*/
#line 1650 "dictionary.web"

async bool itemcmp(pr DICT_NODE**x,pr DICT_NODE**y)
{
return(*x)->key<(*y)->key;
}

/*:34*/
#line 1135 "dictionary.web"

/*29:*/
#line 1429 "dictionary.web"

sync DICT_NODE*insert_nodes(pr DICT_NODE*node,pr DICT_NODE*nextnode,
sh DICT_NODE*dictionary,pr bool leader)
{
sh DICT_NODE*newchildren[6];
pr DICT_NODE*father;
sh int no_of_children,no_of_leaves,nc,j;

#ifdef TESTING
pprintf("Group %d size %d nextnode %d father %d leader %d\n",
MY,groupsize(),nextnode->key,nextnode->father->key,leader);
#endif
if(leader){
father= nextnode->father;
for(nc= 0;nc<father->no_of_children;nc++){
newchildren[2*nc]= NULL;
newchildren[2*nc+1]= father->children[nc];
}
}


newchildren[2*nextnode->child_no]= node;

farm{
if(leader){

no_of_children= 2*father->no_of_children;j= 0;
for(nc= 0;nc<no_of_children;nc++)
if(newchildren[nc]!=NULL)newchildren[j++]= newchildren[nc];
no_of_children= j;
if(no_of_children>father->no_of_children)
/*30:*/
#line 1480 "dictionary.web"

{
#ifdef TESTING
pprintf("Inserting %d children\n",no_of_children);
#endif
j= 0;
if(no_of_children>MAX_CHILDREN){

if(father==dictionary){


node= (DICT_NODE*)malloc(sizeof(DICT_NODE));
node->father= NULL;
node->no_of_children= 1;
node->LEFT= father;
node->key= father->key;
node->no_of_leaves= father->no_of_leaves;

father->child_no= 0;
father->father= node;
#ifdef TESTING
pprintf("Created new root with key %d and %d leaves\n",
node->key,node->no_of_leaves);
#endif
}


node= (DICT_NODE*)malloc(sizeof(DICT_NODE));
node->no_of_children= no_of_children/2;
no_of_leaves= 0;
node->father= NULL;
for(nc= 0;nc<node->no_of_children;nc++){
node->children[nc]= newchildren[nc];
no_of_leaves+= newchildren[nc]->no_of_leaves;
newchildren[nc]->child_no= nc;
newchildren[nc]->father= node;
}
node->key= newchildren[nc-1]->key;
node->no_of_leaves= no_of_leaves;
no_of_children= no_of_children-nc;j= nc;
#ifdef TESTING
pprintf("Created node with key %d\n",node->key);
#endif
}else node= NULL;

father->no_of_children= no_of_children;
no_of_leaves= 0;
for(nc= 0;nc<no_of_children;nc++,j++){
father->children[nc]= newchildren[j];
no_of_leaves+= newchildren[j]->no_of_leaves;
newchildren[j]->child_no= nc;
newchildren[j]->father= father;
}
father->no_of_leaves= no_of_leaves;
#ifdef TESTING
pprintf("Father %d has %d children and %d leaves\n",
father->key,father->no_of_children,father->no_of_leaves);
#endif
}

/*:30*/
#line 1460 "dictionary.web"

else
/*31:*/
#line 1543 "dictionary.web"

{
no_of_leaves= 0;
for(nc= 0;nc<father->no_of_children;nc++)
no_of_leaves+= father->children[nc]->no_of_leaves;
father->no_of_leaves= no_of_leaves;
#ifdef TESTING
pprintf("Changed leaves to %d for key %d\n",
father->no_of_leaves,father->key);
#endif
}

/*:31*/
#line 1462 "dictionary.web"

}else node= NULL;
}

return node;
}

/*:29*/
#line 1136 "dictionary.web"


#define C_insert 1

sync DICT_NODE*dict_insert(sh DICT_NODE*dictionary,
sh DICT_NODE*item[],sh int n)
{
sh int p= groupsize();

sh int*index,*chainsize;
sh int chains;
sh int m;

sh bool*active,*newchain;
sh bool*pending;

sh bool dict_active;

sh DICT_NODE**oldnodes;

pr DICT_NODE*node,*leaf;

pr int i,j;

pr int mygroup;
pr bool leader;

if(n>dict_size(dictionary)/C_insert)/*33:*/
#line 1618 "dictionary.web"

{
oldnodes= 
(DICT_NODE**)shalloc((2*dict_size(dictionary)+n)*sizeof(DICT_NODE*));
m= dict_flatten(dictionary,oldnodes);
Merge(item,n,oldnodes,m,oldnodes+m,sizeof(DICT_NODE*),FC2(int)itemcmp);
if(ME==0){
n= n+m;
oldnodes= oldnodes+m;
}

index= (int*)shalloc(n*sizeof(int));
if(ME==0)index[0]= 1;
for(i= ME+1;i<n;i+= p)
if(oldnodes[i-1]->key==oldnodes[i]->key)index[i]= 0;else index[i]= 1;
prefix_add(index,n,index,-1);
#ifdef TESTING
if(ME==0)
pprintf("Building new dictionary of size %d from %d\n",index[n-1]+1,n);
#endif
for(i= ME+1;i<n;i+= p)
if(oldnodes[i-1]->key!=oldnodes[i]->key)
oldnodes[index[i]]= oldnodes[i];
if(ME==0)n= index[n-1]+1;
dictionary= dict_build(oldnodes,n);

shallfree();
return dictionary;
}

/*:33*/
#line 1163 "dictionary.web"


oldnodes= (DICT_NODE**)shalloc(n*sizeof(DICT_NODE*));

index= (int*)shalloc(2*n*sizeof(int));
chainsize= index+n;
active= (bool*)shalloc(3*n*sizeof(bool));
newchain= active+n;
pending= newchain+n;

/*24:*/
#line 1219 "dictionary.web"


for(i= ME;i<n;i+= p){
farm{
leaf= dict_closest(dictionary,item[i]->key);
}
if(item[i]->key!=leaf->key){

#ifdef TESTING
pprintf("Insert key %d next to %d\n",item[i]->key,leaf->key);
#endif
oldnodes[i]= leaf;
}else{

#ifdef TESTING
pprintf("Replacing key %d\n",item[i]->key);
#endif
item[i]->father= leaf->father;
item[i]->child_no= leaf->child_no;
item[i]->father->children[item[i]->child_no]= item[i];

oldnodes[i]= NULL;

}
item[i]->no_of_children= 0;
item[i]->no_of_leaves= 1;
}

/*:24*/
#line 1173 "dictionary.web"

/*25:*/
#line 1264 "dictionary.web"


for(i= ME;i<n;i+= p)
if(oldnodes[i]==NULL)index[i]= 0;else index[i]= 1;

prefix_add(index,n,index,-1);
#ifdef TESTING
if(ME==0)pprintf("Inserting %d of %d new nodes\n",index[n-1]+1,n);
#endif
for(i= ME;i<n;i+= p)
if(oldnodes[i]!=NULL){
oldnodes[index[i]]= oldnodes[i];
item[index[i]]= item[i];
}
if(ME==0){
n= index[n-1]+1;
newchain[0]= TRUE;
}


for(i= ME+1;i<n;i+= p)
if(oldnodes[i-1]!=oldnodes[i]||i%p==0)newchain[i]= TRUE;
else newchain[i]= FALSE;


for(i= ME;i<n;i+= p){
chainsize[i]= 1;
active[i]= FALSE;
pending[i]= FALSE;
}

group_prefix_add(chainsize,n,chainsize,newchain,&chains);


for(i= ME+1;i<n;i+= p)
if(oldnodes[i-1]!=oldnodes[i]||i%p==0)newchain[i-1]= TRUE;
else newchain[i-1]= FALSE;

if(ME==0)newchain[n-1]= TRUE;

#ifdef TESTING
if(ME==0)pprintf("Chains %d\n",chains);
#endif

/*:25*/
#line 1174 "dictionary.web"

if(n==0)return dictionary;

/*32:*/
#line 1568 "dictionary.web"

if(item[n-1]->key>oldnodes[n-1]->key){
sh DICT_NODE*node;
sh int k;

if(ME==0){
#ifdef TESTING
pprintf("Replacing %d with %d\n",oldnodes[n-1]->key,item[n-1]->key);
#endif
node= oldnodes[n-1];

item[n-1]->child_no= node->child_no;
item[n-1]->father= node->father;

node= node->father;
node->children[node->no_of_children-1]= item[n-1];
while(node!=NULL){
node->key= node->children[node->no_of_children-1]->key;
node= node->father;
}
node= oldnodes[n-1];oldnodes[n-1]= item[n-1];
k= n-1;
}



for(i= n-1-ME;i>1;i-= p)
if(item[i-1]->key<node->key&&node->key<item[i]->key)k= i;
if(ME==0&&item[0]->key>node->key)k= 0;
#ifdef TESTING
if(ME==0)
pprintf("Moving %d into position %d replacing %d\n",
node->key,k,item[k]->key);
#endif
for(i= n-1-ME;i>k;i-= p){
item[i]= item[i-1];
oldnodes[i]= oldnodes[n-1];
}
if(i==k){
item[i]= node;
item[i]->father= NULL;
oldnodes[i]= oldnodes[n-1];
}
}

/*:32*/
#line 1177 "dictionary.web"


for(i= ME;i<n;i+= p)
{
#ifdef TESTING
if(ME==0)pprintf("Next batch!\n");
#endif
if(ME==0)dict_active= TRUE;

while(oldnodes[i]!=NULL){
farm{/*26:*/
#line 1321 "dictionary.web"

if(newchain[i])
if(chainsize[i]>1){

pr int csize;

csize= chainsize[i]/2;
j= i-csize;
chainsize[j]= chainsize[i]-csize;
chainsize[i]= csize;
if(chainsize[j]>1){

chainsize[j-1]= chainsize[j]-1;
chainsize[j]= 1;

newchain[j-1]= TRUE;
oldnodes[j-1]= item[j];
#ifdef TESTING
pprintf("New chain key %d next to %d at %d size %d\n",
item[j-1]->key,oldnodes[j-1]->key,j-1,chainsize[j-1]);
#endif
}
oldnodes[j]= oldnodes[i];
active[j]= TRUE;
pending[j]= TRUE;
#ifdef TESTING
pprintf("Insert new key %d next to %d remaining %d\n",
item[j]->key,oldnodes[j]->key,chainsize[i]);
#endif
}else active[i]= TRUE;

/*:26*/
#line 1187 "dictionary.web"
}
if(active[i]){
node= oldnodes[i]->father;
/*27:*/
#line 1371 "dictionary.web"

for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)oldnodes[i]->father->GROUP_NO= -1;

leader= FALSE;
for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)
if(oldnodes[i]->father->GROUP_NO==(-1)){
oldnodes[i]->father->GROUP_NO= ME;
leader= TRUE;
}

/*:27*/
#line 1190 "dictionary.web"

mygroup= oldnodes[i]->father->GROUP_NO;
/*28:*/
#line 1388 "dictionary.web"

if(leader)oldnodes[i]->father->PENDING= FALSE;

for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)
if(pending[i])oldnodes[i]->father->PENDING= TRUE;

if(leader)pending[i]= oldnodes[i]->father->PENDING;

/*:28*/
#line 1192 "dictionary.web"

fork(p;MY= mygroup;ME= oldnodes[i]->child_no)
item[i]= insert_nodes(item[i],oldnodes[i],dictionary,leader);
farm{
if(leader){
oldnodes[i]= node;
if(oldnodes[i]->father==NULL){
dictionary= oldnodes[i];
oldnodes[i]= NULL;
dict_active= pending[i];
}
}else oldnodes[i]= NULL;
}
}
}
}
return dictionary;
}

/*:23*//*35:*/
#line 1732 "dictionary.web"

/*36:*/
#line 1803 "dictionary.web"

sync DICT_NODE*delete_leaves(pr DICT_NODE*node,pr bool leader)
{
sh bool keepnode[MAX_CHILDREN];
pr int i,nc;
pr DICT_NODE*father;
sh DICT_NODE*lastnode;

father= node->father;
#ifdef TESTING
pprintf("Group %d size %d node %d father %d\n",
MY,groupsize(),node->key,father->key);
#endif
if(leader)
for(nc= 0;nc<father->no_of_children;nc++)keepnode[nc]= TRUE;


if(node->child_no==father->no_of_children-1)lastnode= node;

keepnode[node->child_no]= FALSE;

farm{
if(leader){

i= 0;
for(nc= 0;nc<father->no_of_children;nc++)
if(keepnode[nc]){
father->children[i]= father->children[nc];
father->children[i]->child_no= i;
i++;
}
if(i==0){


father->children[0]= lastnode;
lastnode->child_no= 0;
i= 1;
}else lastnode= NULL;
father->no_of_children= i;
father->no_of_leaves= i;
father->key= father->children[i-1]->key;

node= father;
}
}
if(!leader&&node!=lastnode)node= NULL;

return node;
}

/*:36*//*37:*/
#line 1862 "dictionary.web"

sync DICT_NODE*delete_children(pr DICT_NODE*node,pr bool leader)
{
if(leader)farm{
node= node->father;
/*38:*/
#line 1893 "dictionary.web"

{
pr DICT_NODE*grandchild;
pr int no_of_leaves;
pr int c,cc,gc,nc,newnc,newgc;

#ifdef TESTING
pprintf("Compacting %d children and %d leaves of node %d\n",
node->no_of_children,node->no_of_leaves,node->key);
#endif
newnc= 0;
newgc= node->children[newnc]->no_of_children;
node->no_of_leaves= node->children[newnc]->no_of_leaves;

for(cc= 1;cc<node->no_of_children;cc++){
node->no_of_leaves+= node->children[cc]->no_of_leaves;
if(newgc<2){

nc= node->children[cc]->no_of_children;gc= 0;
no_of_leaves= 0;
while(newgc<2&&gc<nc){
grandchild= node->children[cc]->children[gc++];
grandchild->father= node->children[newnc];
grandchild->child_no= newgc;
no_of_leaves+= grandchild->no_of_leaves;
node->children[newnc]->children[newgc++]= grandchild;
}
node->children[newnc]->key= grandchild->key;
node->children[newnc]->no_of_children= newgc;
node->children[newnc]->no_of_leaves+= no_of_leaves;



c= 0;
while(gc<nc){
grandchild= node->children[cc]->children[gc++];
grandchild->child_no= c;
node->children[cc]->children[c++]= grandchild;
}
node->children[cc]->no_of_children= c;
node->children[cc]->no_of_leaves-= no_of_leaves;
}
if(newgc>=2){

newgc= node->children[++newnc]->no_of_children;
}
}



if(newgc==0)newnc--;
if(newnc>0&&0<newgc&&newgc<2){

nc= node->children[newnc-1]->no_of_children;
if(nc+newgc<=MAX_CHILDREN){

newnc--;
no_of_leaves= 0;
node->children[newnc]->no_of_children= nc+newgc;
while(newgc>0){
grandchild= node->children[newnc+1]->children[--newgc];
grandchild->father= node->children[newnc];
grandchild->child_no= nc+newgc;
no_of_leaves+= grandchild->no_of_leaves;
node->children[newnc]->children[nc+newgc]= grandchild;
}
node->children[newnc]->key= node->children[newnc+1]->key;
node->children[newnc]->no_of_leaves+= no_of_leaves;
}else{


nc= (newgc+nc)/2;
node->children[newnc]->no_of_children= nc;
while(newgc>0){
grandchild= node->children[newnc]->children[--newgc];
node->children[newnc]->children[--nc]= grandchild;
grandchild->child_no= nc;
}
no_of_leaves= 0;
newgc= node->children[newnc-1]->no_of_children;
while(nc>0){
grandchild= node->children[newnc-1]->children[--newgc];
no_of_leaves+= grandchild->no_of_leaves;
grandchild->father= node->children[newnc];
node->children[newnc]->children[--nc]= grandchild;
grandchild->child_no= nc;
}
node->children[newnc-1]->no_of_children= newgc;
node->children[newnc-1]->no_of_leaves-= no_of_leaves;
node->children[newnc-1]->key= 
node->children[newnc-1]->children[newgc-1]->key;
node->children[newnc]->no_of_leaves+= no_of_leaves;
}
}
newnc++;
cc= node->no_of_children;node->no_of_children= newnc;
node->key= node->children[newnc-1]->key;

#ifdef TESTING
pprintf("New children %d and %d leaves of node %d\n",
newnc,node->no_of_leaves,node->key);
#endif


for(;newnc<cc;newnc++){

}
}

/*:38*/
#line 1867 "dictionary.web"

}else node= NULL;
return node;
}

/*:37*/
#line 1733 "dictionary.web"


#define C_delete 1

sync DICT_NODE*dict_delete(sh DICT_NODE*dictionary,
sh DICT_NODE*oldnodes[],sh int n)
{
sh int p= groupsize();

pr DICT_NODE*leaf;
pr bool leader;
pr int i,j,k;

if(n==0)return dictionary;
if(n>dict_size(dictionary)/C_delete)/*40:*/
#line 2029 "dictionary.web"

{
sh DICT_NODE**item;
sh int*index;
sh int m;

item= 
(DICT_NODE**)shalloc(dict_size(dictionary)*sizeof(DICT_NODE*));
index= (int*)shalloc(dict_size(dictionary)*sizeof(int));

m= dict_flatten(dictionary,item);

for(i= ME;i<n;i+= p)oldnodes[i]->father= NULL;
for(i= ME;i<m;i+= p)
if(item[i]->father==NULL)index[i]= 0;else index[i]= 1;
prefix_add(index,m,index,-1);
for(i= ME;i<m;i+= p)
if(item[i]->father==NULL)
item[index[i]]= item[i];
m= index[m-1]+1;
dictionary= dict_build(item,m);

return dictionary;
}

/*:40*/
#line 1747 "dictionary.web"


for(i= ME;i<n;i+= p){
#ifdef TESTING
if(ME==0)pprintf("Next batch!\n");
#endif
while(oldnodes[i]!=NULL){
/*27:*/
#line 1371 "dictionary.web"

for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)oldnodes[i]->father->GROUP_NO= -1;

leader= FALSE;
for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)
if(oldnodes[i]->father->GROUP_NO==(-1)){
oldnodes[i]->father->GROUP_NO= ME;
leader= TRUE;
}

/*:27*/
#line 1754 "dictionary.web"

if(oldnodes[i]->no_of_children==0){


fork(p;
MY= oldnodes[i]->father->GROUP_NO;
ME= oldnodes[i]->child_no)
oldnodes[i]= delete_leaves(oldnodes[i],leader);
}else{


fork(p;
MY= oldnodes[i]->father->GROUP_NO;
ME= oldnodes[i]->child_no)
oldnodes[i]= delete_children(oldnodes[i],leader);
}
/*39:*/
#line 2007 "dictionary.web"

if(oldnodes[i]!=NULL&&oldnodes[i]->father==NULL){

if(dictionary->no_of_children==1){


dictionary= dictionary->children[0];

dictionary->father= NULL;
#ifdef TESTING
pprintf("Removed old root\n");
#endif
}
oldnodes[i]= NULL;
}

/*:39*/
#line 1770 "dictionary.web"




if(oldnodes[i]!=NULL&&oldnodes[i]->no_of_children!=0){


/*27:*/
#line 1371 "dictionary.web"

for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)oldnodes[i]->father->GROUP_NO= -1;

leader= FALSE;
for(j= 0;j<oldnodes[i]->father->no_of_children;j++)
if(j==oldnodes[i]->child_no)
if(oldnodes[i]->father->GROUP_NO==(-1)){
oldnodes[i]->father->GROUP_NO= ME;
leader= TRUE;
}

/*:27*/
#line 1777 "dictionary.web"

fork(p;
MY= oldnodes[i]->father->GROUP_NO;
ME= oldnodes[i]->child_no)
oldnodes[i]= delete_children(oldnodes[i],leader);
/*39:*/
#line 2007 "dictionary.web"

if(oldnodes[i]!=NULL&&oldnodes[i]->father==NULL){

if(dictionary->no_of_children==1){


dictionary= dictionary->children[0];

dictionary->father= NULL;
#ifdef TESTING
pprintf("Removed old root\n");
#endif
}
oldnodes[i]= NULL;
}

/*:39*/
#line 1782 "dictionary.web"

}
}
}

return dictionary;
}

/*:35*/
