/* PAD example. October 1996 */

/* FORK libraries */
#include <fork.h>
#include <io.h>

/* PAD sublibraries */
#include "pad.h"
#include "prefix.h"
#include "merge.h"
#include "list.h"
#include "tree.h"

/* define an associative function for segmented prefix sums */
struct mpair {
  int e; /* element */
  int m; /* marker */
};
async void addmpair(struct mpair *x, struct mpair *y, struct mpair *z)
{
  if (z->m==1) {
    x->e = z->e; /* start of new group */
    x->m = 1;
  } else {
    x->e = y->e+z->e; /* within group: add */
    x->m = y->m;
  }
}

void main(void)
{
  start { /* collect all processors to execute synchronously */
    sh int p ;
    sh int n, m;
    pr int i;
    sh int *x;
    sh int *marker;

    sh struct mpair *a;

    sh TREE *t;
    pr TREE_NODE *v;
    sh LIST *eulerlist;
    sh RANGE *lcaquery;
    sh int deg, ch;

    p = __STARTED_PROCS__ ; /* get number of available processors */

    seq {
      n = 100; m = 25;
    }
    x = (int *) shalloc(n*sizeof(int));
    marker = (int *) shalloc(n*sizeof(int));
    a = (struct mpair *) shalloc(n*sizeof(struct mpair));

    /* segment, and compute segmented prefix sums */
    aritseq(x,n,1,1);
    prefix_add(x,n,x,0);
    par_array(i,n) marker[i] = 0;
    par_array(i,n) if (x[i]<n) marker[x[i]] = 1;
    par_array(i,n) {
      a[i].e = 1;
      a[i].m = marker[i];
    }
    Prefix(a,n,sizeof(struct mpair),PR3 addmpair);
    par_array(i,n) pprintf("%d: (%d,%d)\n",i,a[i].e,a[i].m);

    /* simple examples of tree computations */
    seq {
      n = 50; deg = 5; ch = 3;
    }
    seq pprintf("Generating small complete tree\n");
    t = complete_tree(n,deg,ch);
    seq pprintf("Actual treesize n=%d m=%d\n",tree_n(t),tree_e(t));

    eulertour(t); /* make an Euler tour */
    eulerlist = (LIST *) shalloc(tree_e(t)*sizeof(LIST)); /* allocate */
    seq pprintf("Rooting tree\n");
    tree_root(t,NODE(0,t),eulerlist);
    par_nodes(v,t) {
      pprintf("node %d level %d pre %d post %d desc %d\n",
	      NODE_NO(v,t),
	      node_level(v),node_pre(v),node_post(v),node_desc(v));
      if (LEAF(v)) pprintf("leaf %d\n",NODE_NO(v,t));
    }

    /* LCA stuff */
    lcaquery = lca_preprocess(t,eulerlist);
    par_array(i,tree_n(t)-1)
      pprintf("lca(%d,%d) = %d\n",i,i+1,
	      NODE_NO(lca(NODE(i,t),NODE(i+1,t),lcaquery),t));
    tree_height(t,eulerlist);
    par_nodes(v,t)
      pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));
  }
}
