/*46:*/
#line 1600 "trees.web"

#include <fork.h>
#include <io.h>
#include "pad.h"
#include "prefix.h"
#include "list.h"
#include "tree.h"

void main(void)
{
start{
sh int p;
sh int n,m;
sh int deg,ch;

pr TREE_NODE*v;
pr TREE_EDGE*e;

sh TREE_NODE*root;
sh LIST*eulerlist;
sh RANGE*lcaquery;
sh TREE*t,*ct;
pr int i;
pr int s;

p= __STARTED_PROCS__;

n= 50;deg= 5;ch= 3;
seq pprintf("Generating small complete tree\n");

t= complete_tree(n,deg,ch);
seq pprintf("Actual treesize n=%d m=%d\n",t->n,t->m);


eulertour(t);

eulerlist= (LIST*)shalloc(t->m*sizeof(LIST));
seq pprintf("Rooting tree\n");
s= getct();
tree_root(t,NODE(0,t),eulerlist);
s= getct()-s;
par_nodes(v,t){
pprintf("node %d: level %d pre %d post %d desc %d\n",
NODE_NO(v,t),node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,t));
}
seq pprintf("Parallel rooting time %d\n",s);



lcaquery= lca_preprocess(t,eulerlist);
par_array(i,t->n-1)
pprintf("lca(%d,%d) = %d\n",i,i+1,
NODE_NO(lca(NODE(i,t),NODE(i+1,t),lcaquery),t));
par_array(i,t->n)
pprintf("lca(%d,%d) = %d\n",i,t->n-1-i,
NODE_NO(lca(NODE(i,t),NODE(t->n-1-i,t),lcaquery),t));

tree_height(t,eulerlist);
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));



seq pprintf("Copying tree and rooting sequentially\n");
ct= tree_alloc(t->n);
tree_copy(t,ct);
s= getct();
seq seq_dfs(ct,NODE(0,ct));
s= getct()-s;
par_nodes(v,ct){
pprintf("node %d: level %d pre %d post %d desc %d\n",
NODE_NO(v,ct),node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,ct));
}
seq pprintf("Sequential rooting time %d\n",s);


seq pprintf("Rerooting (original tree), root %d\n",30);
eulertour(t);
tree_root(t,NODE(30,t),eulerlist);
par_nodes(v,t){
pprintf("node %d level %d pre %d post %d desc %d\n",
NODE_NO(v,t),node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,t));
}



lcaquery= lca_preprocess(t,eulerlist);
par_array(i,t->n-1)
pprintf("lca(%d,%d) = %d\n",i,i+1,
NODE_NO(lca(NODE(i,t),NODE(i+1,t),lcaquery),t));
par_array(i,t->n)
pprintf("lca(%d,%d) = %d\n",i,t->n-1-i,
NODE_NO(lca(NODE(i,t),NODE(t->n-1-i,t),lcaquery),t));

tree_height(t,eulerlist);
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));


seq pprintf("Tree contraction for binary trees.\n");
n= 50;deg= 2;ch= 1;
t= complete_tree(n,deg,ch);
seq pprintf("Actual treesize n=%d m=%d\n",t->n,t->m);
eulertour(t);

eulerlist= (LIST*)shalloc(t->m*sizeof(LIST));
seq pprintf("Rooting tree\n");
tree_root(t,t->nodes,eulerlist);
par_nodes(v,t){
pprintf("node %d sib %d: level %d pre %d post %d desc %d\n",
NODE_NO(v,t),NODE_NO(v->sibling,t),
node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,t));
}


bintree_height(t,eulerlist);
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));

n= 50;deg= 2;ch= 2;
t= complete_tree(n,deg,ch);
seq pprintf("Actual treesize n=%d m=%d\n",t->n,t->m);
eulertour(t);

eulerlist= (LIST*)shalloc(t->m*sizeof(LIST));
seq pprintf("Rooting tree\n");
tree_root(t,t->nodes,eulerlist);
par_nodes(v,t){
pprintf("node %d sib %d: level %d pre %d post %d desc %d\n",
NODE_NO(v,t),NODE_NO(v->sibling,t),
node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,t));
}


seq pprintf("Contracting\n");
s= getct();
bintree_height(t,eulerlist);
s= getct()-s;
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));
seq pprintf("Contraction time %d\n",s);
s= getct();
lcaquery= lca_preprocess(t,eulerlist);
tree_height(t,eulerlist);
s= getct()-s;
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));
seq pprintf("Height computation (range query) time %d\n",s);


seq{
pprintf("Deleting nodes and subtree, compacting and rooting again.\n");
t->nodes[25].firstedge= NULL;
t->nodes[26].firstedge= NULL;
t->nodes[35].firstedge= NULL;
t->nodes[36].firstedge= NULL;
}
mark_subtree(t,NODE(7,t),eulerlist);
tree_compact(t);
seq pprintf("New treesize n=%d m=%d\n",t->n,t->m);
eulertour(t);
seq pprintf("Rooting\n");
tree_root(t,NODE(0,t),eulerlist);
par_nodes(v,t){
pprintf("node %d sib %d: level %d pre %d post %d desc %d\n",
NODE_NO(v,t),NODE_NO(v->sibling,t),
node_level(v),node_pre(v),node_post(v),node_desc(v));
if(LEAF(v))pprintf("leaf %d\n",NODE_NO(v,t));
}

seq pprintf("Contracting\n");
s= getct();
bintree_height(t,eulerlist);
s= getct()-s;
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));
seq pprintf("Contraction time %d\n",s);

s= getct();
lcaquery= lca_preprocess(t,eulerlist);
tree_height(t,eulerlist);
s= getct()-s;
par_nodes(v,t)pprintf("height %d: %d\n",NODE_NO(v,t),node_height(v));
seq pprintf("Height computation (range query) time %d\n",s);
}
}

/*:46*/
